#ifndef RAWCONFIGURATIONBLOCK_H
#define RAWCONFIGURATIONBLOCK_H

#include "DAQDataFormats/RawConfigurationHeader.h"
#include "DAQDataFormats/RawConfigurationTail.h"
#include "DAQDataFormats/RawConfigurationNamespace.h"
#include "DAQDataFormats/RawConfigurationSystemID.h"

namespace daqdataformats {

#define DEFAULT_CONFIGURATIONBLOCK_SIZE 1000 /* Maximum MilliSlice Buffer Size  */

  class RawConfigurationBlock : public RawDAQData {
    
  public:
    // Constructors
    RawConfigurationBlock (uint32_t buffersize = DEFAULT_CONFIGURATIONBLOCK_SIZE); //! The default constructor
    RawConfigurationBlock (const RawConfigurationBlock& copy_in);                  //! Deep copy constructor
    RawConfigurationBlock (void* configurationblock_buffer, uint32_t buffersize = DEFAULT_CONFIGURATIONBLOCK_SIZE); //! Shallow copy constructor
    /// Destructor
    ~RawConfigurationBlock();

    /// Initialization of Data members
    void init();
    
    /// Deep Copy
    void copy(const RawConfigurationBlock& copy_in);

    /// Method to close the Configuration Header. Sets the lower markers, the CRC.
    bool close();

    /// Add a DAQ system to the Configuration Header
    void addSystem(uint32_t sysID, uint32_t sysVersion);

    /// Read method is for reading the whole RawConfigurationBlock
    void* readData(void* configurationblock_buffer);
    // Suplementary reading methods
    void* readData(void* buffer, uint32_t size);
    void* readData(void* buffer, uint32_t size, uint32_t _Buffer_start_position);

    /// Size of MilliSlice _Buffer in words
    uint32_t sizeofdata() const;

    /// Print the MilliSlice structure
    void print() const;


    // Getters
    RawConfigurationHeader*    getHeader();
    RawConfigurationTail*      getTail();
    RawConfigurationNamespace* getNamespace();
    RawConfigurationSystemID*  getSystemID();

    uint32_t getMarkerHi1()                        const;
    uint32_t getMarkerHi2()                        const;
    uint32_t getMarkerLo1()                        const;
    uint32_t getMarkerLo2()                        const;
    uint32_t getConfigurationHeaderVersion()       const;
    uint32_t getDetectorID()                       const;
    uint32_t getConfigurationBlockSize()           const;
    uint32_t getResourceManagerID()                const;
    uint32_t getRunNumber()                        const;
    uint32_t getRunType()                          const;
    uint32_t getNameSize()                         const;
    uint32_t getPartition()                        const;
    uint32_t getSubrun()                           const;
    uint32_t getSystemNumber()                     const;
    uint32_t getSystemVersionIDX()                 const;
    uint32_t getDAQSystemVersion(uint32_t isystem) const;
    uint32_t getDAQSystemID(uint32_t isystem)      const;
    uint32_t getFreeMemory()                       const;   //! Free memory left in the buffer.
    uint32_t getNamespace( std::string& ); //! returns the number of words

    
    // Setters
    bool setConfigurationHeaderVersion (uint32_t);
    bool setDetectorID                 (uint32_t);
    bool setResourceManagerID          (uint32_t);
    bool setRunNumber                  (uint32_t);
    bool setRunType                    (uint32_t);
    bool setPartition                  (uint32_t);
    bool setSubrun                     (uint32_t);
    bool setSystemVersionIDX           (uint32_t);
    bool setNamespace( std::string& ); 

    bool setBufferSource (void* pointer); //!< Set The Buffer pointer to a given location. Don't use it if you are not sure
    bool resetBufferSource();             //!< Reset the class to use its internal buffer
    bool clear();                         //!< Clear the buffer
    bool reset();                         //!< Reset the class data
    bool isInternalBuffer() const;        //!< Is the buffer internal?

  private:

    RawConfigurationHeader    _Header;    //! Header. Embedded into the InternalBuffer of the RawConfigurationBlock
    RawConfigurationTail      _Tail;      //! Tail. Uses its own Internal Buffer.
    RawConfigurationNamespace _Namespace; //! Configuration namespace. Uses its own Internal Buffer. 
    RawConfigurationSystemID  _SystemID;  //! Configuration namespace. Uses its own Internal Buffer.
                                          //! Both _Tail, _Namespace and _SystemID need ConfigurationBlock to be closed
                                          //! in the end to be written out in
                                          //! the InternalBuffer of the RawConfigurationBlock.


     /// Maximum size of the current buffer
    uint32_t                _max_buffer_size;

    /// The actual internal buffer (not a pointer).
    /// Used only if the constructors with the default size are used
    std::vector <uint32_t>  _InternalBuffer;


  };
  
  
} // END NAMESPCE daqdataformats
#endif /* RAWCONFIGURATIONBLOCK_H */
