#ifndef RAWCONFIGURATIONHEADER_H
#define RAWCONFIGURATIONHEADER_H

#include "DAQDataFormats/RawDAQData.h"

namespace daqdataformats {

#define RAWCONFIGURATIONHEADER_SIZE 7


    namespace rawconfhdr {

        // Configuration header delimiter
        static const uint32_t Marker_value[2] = {0x464E4F43, 0x54525453};

        enum RunHeaderMASKS{
            MASK_MarkerHi1                  = 0xFFFFFFFF, //! Marker Mask upper word 1      (32 bits)
            MASK_MarkerHi2                  = 0xFFFFFFFF, //! Marker Mask upper word 2      (32 bits)
            MASK_ConfigurationHeaderVersion = 0x00000FFF, //! Configuration header version  (12 bits)
            MASK_DetectorID                 = 0x00007000, //! Detector ID                   (3 bits)
            MASK_ConfigurationBlockSize     = 0xFFFF0000, //! Size of the Configuration Block in words (16 bits)
            MASK_ResourceManagerID          = 0x0000FFFF, //! Recource manager identifier   (16 bits)
            MASK_RunNumber                  = 0x00FFFFFF, //! Run number               (24 bits)
            MASK_RunType                    = 0xFF000000, //! Run type                 (8 bits)
            MASK_NameSize                   = 0x0000FFFF, //! Namespace size           (16 bits)
            MASK_Partition                  = 0x00FF0000, //! Partition identifier that Run Control and Data Logger belong to (8 bits)
            MASK_Subrun                     = 0x3F000000, //! Subrun number            (6 bits)
            MASK_SystemNumber               = 0x0000FFFF, //! Number of systems        (16 bits)
            MASK_SystemVersionIDX           = 0xFFFF0000, //! System versionindex key  (16 bits)

        };

        enum RunHeaderSHIFTS{
            SHIFT_MarkerHi1                  = 0,  //! Marker Mask upper word (32 bits)
            SHIFT_MarkerHi2                  = 0,  //! Marker Mask lower word (32 bits)
            SHIFT_ConfigurationHeaderVersion = 0,  //! Configuration header version  (12 bits)
            SHIFT_DetectorID                 = 12, //! Detector ID 
            SHIFT_ConfigurationBlockSize     = 16, //! Size of the Configuration Block in words (16 bits)
            SHIFT_ResourceManagerID          = 0,  //! Recource manager identifier   (16 bits)
            SHIFT_RunNumber                  = 0,  //! Run number             (24 bits)
            SHIFT_RunType                    = 24, //! Run type               (8 bits)
            SHIFT_NameSize                   = 0,  //! Detector configuration identifier (16 bits)
            SHIFT_Partition                  = 16, //! Partition identifier that Run Control and Data Logger belong to (8 bits)
            SHIFT_Subrun                     = 24, //! Subrun number          (6 bits)
            SHIFT_SystemNumber               = 0,  //! Number of systems
            SHIFT_SystemVersionIDX           = 16, //! System versionindex key
        };

        enum RunHeaderWORDS{
            WORD_MarkerHi1                  = 0, //! Marker Mask upper word (32 bits)
            WORD_MarkerHi2                  = 1, //! Marker Mask lower word (32 bits)
            WORD_ConfigurationHeaderVersion = 2, //! Configuration header version  (12 bits)
            WORD_DetectorID                 = 2,
            WORD_ConfigurationBlockSize     = 2, //! Size of the Configuration Block in words (16 bits)
            WORD_ResourceManagerID          = 3, //! Recource manager identifier   (16 bits)
            WORD_RunNumber                  = 4, //! Run number             (24 bits)
            WORD_RunType                    = 4, //! Run type               (8 bits)
            WORD_NameSize                   = 5, //! Detector configuration identifier (16 bits)
            WORD_Partition                  = 5, //! Partition identifier that Run Control and Data Logger belong to (8 bits)
            WORD_Subrun                     = 5, //! Subrun number          (6 bits)
            WORD_SystemNumber               = 6, //! Number of systems
            WORD_SystemVersionIDX           = 6, //! System versionindex key
        };
    }// end of namespace rawconfhdr


  class RawConfigurationHeader : public RawDAQData {
    
  public:
    // Constructors
    RawConfigurationHeader ();                                      //! The default constructor
    RawConfigurationHeader (const RawConfigurationHeader& copy_in); //! Deep copy constructor
    RawConfigurationHeader (void* configurationheader_buffer);      //! Shallow copy constructor
    /// Destructor
    ~RawConfigurationHeader();

    /// Initialization of Data members
    void init();
    
    /// Deep Copy
    void copy(const RawConfigurationHeader& copy_in);

    /// Size of MilliSlice _Buffer in words
    uint32_t sizeofdata() const;

    /// Print the MilliSlice structure
    void print() const;


    // Getters
    uint32_t getMarkerHi1()                        const;
    uint32_t getMarkerHi2()                        const;
    uint32_t getConfigurationHeaderVersion()       const;
    uint32_t getDetectorID()                       const;
    uint32_t getConfigurationBlockSize()           const;
    uint32_t getResourceManagerID()                const;
    uint32_t getRunNumber()                        const;
    uint32_t getRunType()                          const;
    uint32_t getNameSize()                         const;
    uint32_t getPartition()                        const;
    uint32_t getSubrun()                           const;
    uint32_t getSystemNumber()                     const;
    uint32_t getSystemVersionIDX()                 const;
    
    // Setters
    bool setMarkerHi1                  (uint32_t);
    bool setMarkerHi2                  (uint32_t);
    bool setConfigurationHeaderVersion (uint32_t);
    bool setDetectorID                 (uint32_t);
    bool setConfigurationBlockSize     (uint32_t);
    bool setResourceManagerID          (uint32_t);
    bool setRunNumber                  (uint32_t);
    bool setRunType                    (uint32_t);
    bool setNameSize                   (uint32_t);
    bool setPartition                  (uint32_t);
    bool setSubrun                     (uint32_t);
    bool setSystemNumber               (uint32_t);
    bool setSystemVersionIDX           (uint32_t);
    bool advanceSystemNumber();
    
    bool setBufferSource (void* pointer); //!< Set The Buffer pointer to a given location. Don't use it if you are not sure
    bool resetBufferSource();             //!< Reset the class to use its internal buffer
    bool clear();                         //!< Clear the buffer
    bool reset();                         //!< Reset the class data
    bool isInternalBuffer() const;        //!< Is the buffer internal?

  private:

      /// The actual internal buffer (not a pointer)
      uint32_t  _InternalBuffer[RAWCONFIGURATIONHEADER_SIZE];

  };
  
  
} // END NAMESPCE daqdataformats
#endif /* RAWCONFIGURATIONHEADER_H */
