#ifndef RAWDAQDATA_BASE_H
#define RAWDAQDATA_BASE_H

#include "DAQDataFormats/version.h"   // Package Version Number

#include <sys/types.h>		/* System based data types */
#include <iostream>             /* Iostream for printing   */
#include <assert.h>             //assertion
#include <stdio.h>

#include <DAQDataFormats/BitFields.h> // most classes will use this
#include <boost/crc.hpp> // Boost library for CRC
/// CRC-32C (Castagnoli) iSCSI polynomial (See A.Norman, NOvA-doc-DB-4619)
typedef boost::crc_optimal<32, 0x1EDC6F41, 0x0, 0x0, true, true> nova_crc_32_type;


/*! 
  The RawDAQData class is a virtual class definition that
  all the NOVA DAQ data format classes inherit from.
  
  All RawDAQData class decendants must impliment:
  
  A writeData method that writes the objects data in a 
  packed format to a memory location.

  A readData method that reads the objects data in a 
  packed format from a memory location.

  and

  A sizeofdata method that returns the native size of the
  packed data object in 32bit words

*/
namespace daqdataformats {

    typedef enum {
        rds_error = -1,
        rds_incomplete,
        rds_complete,
    } readDataStatus;


class RawDAQData {

 public:
  RawDAQData();  //!< Default Constructor.
  virtual ~RawDAQData(); //!< Default Destructor.

  //---------------------------------------------------------
  // Data I/O Methods

  // General read method that read the raw formated data block from the target memory
  virtual void*   readData(void* buffer ); //!< Read Data from target memory location. Return is the address of the end of the read. NULL on error.
  virtual readDataStatus readData(int fd); //!< Read Data from target file descriptor into internal buffer where size is known

  // General write method that writes the raw formated data block to the target memory
  void*   writeData(void* buffer);			  //!< Write Data to target memory location. Return is the address of the end of the write. NULL on error.
  ssize_t writeData(int fd);				  //!< Write Data to target file descriptor. Return is the number of bytes written. -1 on error.

  //---------------------------------------------------------
  // Data Display Methods
  void printBuffer();		                              //!< Displays the data _Buffer in bits
  void printWord (uint32_t iword, bool is_bynary_printing) const;   //! Print the i-th word of _Buffer, is_bynary_printing: true = nynary, false = HEX

  virtual void print() const; //! General method for printing the content of the Buffer
                        //! Should be overloaded
  //---------------------------------------------------------
  // Data Validation Methods

  /// Calculate CRC of _Buffer (after skipping last skip_last_words words) using CRC32c algorythm with 0x8F6E37A0 polynomial.
  /// Usually use skip_last_words=1 as the _Buffer has an CRC in the end
  virtual uint32_t calculateCheckSum(int skip_last_words);

  /// Add the CRC into the end of fBuffer. CRC of all but the last words of the _Buffer
  virtual void addCheckSum();


  /// Size of the current data block in words
  virtual uint32_t sizeofdata() const = 0; //!< Sizeof method for the current data block.  Returns the size of the data block in 32bit words.

  /// Get method for the pointer to the internal buffers
  void* getBuffer() const;		/* This is evil.  You shouldn't expose the buffer point. However, we allow it. */

  int getMajorVersion(){return major_version;}
  int getMinorVersion(){return minor_version;}

 protected:

  int major_version;
  int minor_version;

  ///All data formats need to have _Buffer and a Data structure. _Buffer will be common.
  // void* _BUFF_;
  void* _Buffer;                //!< The internal Buffer for each data format
                                //!< Generalized pointer to the buffer holding the packed data
                                //!< This pointer is used in performing generalized read/write calls
                                //!< and should be set during initialization
  
  nova_crc_32_type _CRC32;      //!< CRC-32C object. Used when producing CRC
 };


}// end of daqdataformats namespace
#endif /* RAWDAQDATA_BASE_H */
