#ifndef RAWEVENT_H
#define RAWEVENT_H

#include "DAQDataFormats/RawEventHeader.h"
#include "DAQDataFormats/RawEventTail.h"
#include "DAQDataFormats/RawDataBlock.h"
#include "DAQDataFormats/RawTrigger.h"

namespace daqdataformats {

//    #define MAXIMUM_RAWEVENT_SIZE 262144 /* Maximum Microslice Buffer Size 64k */
    #define DEFAULT_RAWEVENT_SIZE 262144 /* Maximum Microslice Buffer Size 64k */


  class RawEvent : public RawDAQData {
    
  public:
    /// Constructor. Makes an empty MilliSlice
    RawEvent (uint32_t buffersize = DEFAULT_RAWEVENT_SIZE);                       //! The default constructor
    RawEvent (const RawEvent& copy_in);                                           //! Deep copy constructor
    RawEvent (void* rawevent_buffer, uint32_t buffersize = DEFAULT_RAWEVENT_SIZE);//! Shallow copy constructor
    /// Destructor
    ~RawEvent();

    /// Initialization of Data members
    void init();

    /// Deep Copy
    void  copy(const RawEvent& copy_in);

    // Next event. Advance the event number and clear the DataBlocks and CRC
    void nextEvent();

    ///Add a MicroSlice into the MilliSlice changing the MilliSlice Header appropriately
    /// If check_mc_flag = true, the method checks whether the DataBlock is MC generated
    /// and adds an appropriate flag to the Event Header.
    void addDataBlock(void* datablock_buffer, bool check_mc_flag = true);
    ///Add a MicroSlice of into the MilliSlice changing the MilliSlice Header appropriately
    /// If check_mc_flag = true, the method checks whether the DataBlock is MC generated
    /// and adds an appropriate flag to the Event Header.
    void addDataBlock(void* datablock_buffer, uint32_t datablock_size, bool check_mc_flag = true);

    /// Adds the tail and CRC
    /// if produce_crc=flase, don't calculate CRC.
    void close(bool produce_crc = true);

    /// Read method is for reading the whole RawEvent
    void* readData(void* rawevent_buffer);
    // Suplementary reading methods
    void* readData(void* buffer, uint32_t size);
    void* readData(void* buffer, uint32_t size, uint32_t _Buffer_start_position);


    /// Size of MilliSlice _Buffer in words
    uint32_t sizeofdata() const;

    /// Print the MilliSlice structure
    void print() ;

    // Getters
    RawEventHeader* getHeader();
    RawEventTail*   getTail();
    RawTrigger*     getTrigger();
    RawDataBlock*   getFloatingDataBlock();
    uint32_t        getHeaderMarkerLeft()  const;
    uint32_t        getHeaderMarkerRight() const;
    uint32_t        getVersion()           const;
    uint32_t        getDataLoggerID()      const;
    uint32_t        getRunNumber()         const;
    uint32_t        getRunType()           const;
    uint32_t        getConfigIDX()         const;
    uint32_t        getPartition()         const;
    uint32_t        getSubrun()            const;
    uint32_t        getEventNumberLo()     const;
    uint32_t        getEventNumberHi()     const;
    uint64_t        getEventNumber()       const;
    uint32_t        getEventType()         const;
    uint32_t        getEventSize()         const;
    uint32_t        getDataBlockNumber()   const;
    bool            isEventIncomplete()    const;
    bool            getMonteCarloFlag()    const;
    bool            isCRCCalculationUsed() const;
    uint32_t        getFreeMemory()        const; //! Free memory left in the buffer.



    // Setters
    //Different setters for the trigger
    bool setTrigger           (RawTrigger&  copyin);
    bool setTrigger           (void*  pointer);

    bool setVersion           (uint32_t);
    bool setDataLoggerID      (uint32_t);
    bool setRunNumber         (uint32_t);
    bool setRunType           (uint32_t);
    bool setConfigIDX         (uint32_t);
    bool setPartition         (uint32_t);
    bool setSubrun            (uint32_t);
    bool setEventNumberLo     (uint32_t);
    bool setEventNumberHi     (uint32_t);
    bool setEventNumber       (uint64_t);
    bool setEventType         (uint32_t);
    bool setEventIncomplete   (bool);
    bool setMonteCarloFlag    (bool);
    bool setFloatingDataBlock (uint32_t idatablock);//The count starts from 0.

    bool setBufferSource (void* pointer); //!< Set The Buffer pointer to a given location. Don't use it if you are not sure
    bool resetBufferSource();             //!< Reset the class to use its internal buffer
    bool clear();                         //!< Clear the buffer
    bool reset();                         //!< Reset the class data
    bool isInternalBuffer() const;              //!< Is the buffer internal?

  private:
    /// Check size of the input RawEvent
    bool checkSize(uint32_t size) const;

  private:
    
    RawEventHeader                 _Header;                  //! RawEvent Header
    RawEventTail                   _Tail;                    //! RawEvent Tail
    RawTrigger                     _Trigger;                 //! Trigger
    RawDataBlock                   _FloatingDataBlock;       //! Floating DataBlock. The count starts from 0.
    int                            _FloatingDataBlockNumber; //! Floating DataBlock number
    // Tail will be constructed in the close


    //! Maximum size of the current buffer
    uint32_t              _max_buffer_size;
    /// The actual internal buffer (not a pointer)
    std::vector<uint32_t> _InternalBuffer;

    /// The actual internal buffer (not a pointer)
    //uint32_t  _InternalBuffer[MAXIMUM_RAWEVENT_SIZE];

  };
  
  
} // END NAMESPCE daqdataformats
#endif /* RAWEVENT_H */
