#ifndef RAWMILLISLICE_H
#define RAWMILLISLICE_H

#include "DAQDataFormats/RawMicroSlice.h"
#include "DAQDataFormats/RawMilliSliceHeader.h"
#include "DAQDataFormats/RawMilliSliceIndex.h"

#include <stdarg.h>		// va_start

namespace daqdataformats {

    #define DEFAULT_MILLISLICE_SIZE 262144 /* Maximum MilliSlice Buffer Size  */

  class RawMilliSlice : public RawDAQData {
    
  public:
    // Constructors
    RawMilliSlice (uint32_t bufsize=DEFAULT_MILLISLICE_SIZE, uint32_t tlvl=0); //! The default constructor
    RawMilliSlice (const RawMilliSlice& copy_in);                  //! Deep copy constructor
    RawMilliSlice (void* millislice_buffer, uint32_t buffersize = DEFAULT_MILLISLICE_SIZE); //! Shallow copy constructor
    /// Destructor
    ~RawMilliSlice();

    /// Initialization of Data members
    void init();
    
    /// Deep Copy
    void copy(const RawMilliSlice& copy_in);

    /// Method to close the MilliSlice. Produces the tail and the CRC.
    /// if produce_crc=flase, don't calculate CRC.
    bool close(bool produce_crc = true);

    /// Add some data containing NumberofMicroslices MicroSlices into the MilliSlice changing the MilliSlice Header appropriately
    /// if indexate is true, produce an index for the MicroSlice
    /// If check_mc_flag = true, the method checks whether the microslice is MC generated
    /// and adds an appropriate flag to the MilliSlice Header.
    void addMicroSlice(void* microslice_buffer, bool indexate = false, bool check_mc_flag = true);
    /// Add some data containing NumberofMicroslices MicroSlices into the MilliSlice changing the MilliSlice Header appropriately
    /// If check_mc_flag = true, the method checks whether the microslice is MC generated
    /// and adds an appropriate flag to the MilliSlice Header.
    void addData(void* buffer, uint32_t size, uint32_t NumberofMicroslices, 
                 bool check_mc_flag = true, bool check_on_space = true);

    /// Read method is for reading the whole RawMilliSlice
    void* readData(void* millislice_buffer);
    // Suplementary reading methods
    void* readData(void* buffer, uint32_t size);
    void* readData(void* buffer, uint32_t size, uint32_t _Buffer_start_position);
    virtual readDataStatus readData( int fd );

    /// Size of MilliSlice _Buffer in words
    uint32_t sizeofdata() const;

    /// Print the MilliSlice structure
    void print() const;

    // Getters
    RawMilliSliceHeader* getHeader();                        //! MilliSlice Header reference
    RawMicroSlice*       getFloatingMicroSlice();            //! The current floating MicroSlice reference
    RawMilliSliceIndex*  getMilliSliceIndex();               //! MilliSlice Index reference
    
    uint32_t             getVersion()                 const; //! Get the MilliSlice version
    uint32_t             getMarker()                  const; //! Get the MilliSlice Marker (Delimiter)
    uint32_t             getDestBuffID()              const; //! Get the destination Buffer ID
    uint32_t             getPartition()               const; //! Get the DCM Partition
    uint64_t             getTimeStart()               const; //! MilliSlice Start time
    uint32_t             getTimeStartLow()            const; //! Lower bits of the MilliSlice Start time
    uint32_t             getTimeStartHigh()           const; //! Upper bits of the MilliSlice Start time
    uint32_t             getTimeLength()              const; //! Time Length of the MilliSlice
    bool                 isComplete()                 const; //! Get the Completeness bit
    rawmillihdr::Mode_t  getMode()                    const; //! Get the MilliSlice Mode
    bool                 isConnectionInitialization() const; //! Get the Initialization Connection bit
    bool                 isCloseConnection()          const; //! Get the Closing Connection bit
    uint32_t             getMilliSliceSEQ()           const; //! Get the MilliSlice Sequence
    uint32_t             getFrameSEQ()                const; //! Get the frame 5mus Frame sequence
    uint32_t             getSliceSize()               const; //! Get the size of data outside of the MilliSlice header
    uint32_t             getMicrosliceNumber()        const; //! Get the Number of MicroSlices in the MilliSlice
    uint32_t             getDiblockId()               const; //! Get the Diblock ID
    uint32_t             getDCMId()                   const; //! Get the DCM ID
    uint32_t             getDetId()                   const; //! Get the Detector ID
    uint32_t             getGlobalDCMID()             const; //! Get the Detector ID
    uint32_t             getMilliSliceIndexNumber()   const; //! Get the number of MilliSlice indeces
    uint32_t             getMicroSliceTimeLength_usec() const; //! Get the configured TimeLenght of microslices
    bool                 getMonteCarloFlag()          const; //! Get the Data Present bit of the Header
    bool                 isCRCCalculationUsed()       const; //! Is CRC calculation used?
    bool                 isMilliSliceIndexUsed()      const; //! Is MilliSlice Index EVER used?
    uint32_t             getFreeMemory()              const; //! Free memory left in the buffer.
    
    // Setters
    bool setDiblockId                (uint32_t);            //! Set the Diblock ID
    bool setDCMId                    (uint32_t);            //! Set the DCM ID
    bool setDetId                    (uint32_t);            //! Set the Detector ID
    bool setGlobalDCMID              (uint32_t);            //! Set the Detector ID
    bool setVersion                  (uint32_t);            //! Set the MilliSlice version
    bool setDestBuffID               (uint32_t);            //! Set the destination Buffer ID
    bool setPartition                (uint32_t);            //! Set the DCM Partition
    bool setTimeStart                (uint64_t);            //! MilliSlice Start time
    bool setTimeStartLow             (uint32_t);            //! Lower bits of the MilliSlice Start time
    bool setTimeStartHigh            (uint32_t);            //! Upper bits of the MilliSlice Start time
    bool setTimeLength               (uint32_t);            //! Time Length of the MilliSlice
    bool setComplete                 (bool);                //! Set the MilliSlice Mode
    bool setMode                     (rawmillihdr::Mode_t); //! Set the MilliSlice Mode
    bool setConnectionInitialization (bool);                //! Set the Initialization Connection bit
    bool setCloseConnection          (bool);                //! Set the Closing Connection bit
    bool setMilliSliceSEQ            (uint32_t);            //! Set the MilliSlice Sequence
    bool setMicroSliceTimeLength_usec(uint32_t);            //! Set the MicroSliceTimeLength; param is microseconds
    bool setFloatingMicroSlice       (uint32_t imicro);     //! Set the Floating MicroSlice to the "imicro"-th MicroSlice in the MilliSlice
    bool setMonteCarloFlag           (bool);                //! Set the bit to if SIM flag
    bool setMilliSliceIndexUsed      (bool);                //! Set the bit to if MilliSlice Index ever used
    bool setFrameSEQ                 (uint32_t);

    bool setBufferSource (void* pointer); //!< Set The Buffer pointer to a given location. Don't use it if you are not sure
    bool resetBufferSource();             //!< Reset the class to use its internal buffer
    bool clear();                         //!< Clear the buffer
    bool reset();                         //!< Reset the class data
    bool isInternalBuffer() const;        //!< Is the buffer internal?

  private:
    // The actual internal buffer (not a pointer).
    // Used only if the constructors with the default size are used
    std::vector <uint32_t>  _InternalBuffer;

    RawMilliSliceHeader     _Header;                  //! MilliSlice Header
    RawMilliSliceIndex      _Index;                   //! MilliSlice Index

    RawMicroSlice           _FloatingMicroSlice;      //! Floating MicroSlice
    int                     _FloatingMicroSliceNumber;//! Floating MicroSlice Number

    size_t                  _fd_read_bytes;
    //! Maximum size of the current buffer
    uint32_t                _max_buffer_size;

    uint32_t                _tlvl; // debug "trace level mask"

      void tprintf( unsigned lvl, const char *fmt, ... )
	__attribute__((format(printf, 3, 4))); // g++ note: attributes are not
               			          //  allowed on a function-definition
  };

  
} // END NAMESPCE daqdataformats
#endif /* RAWMILLISLICE_H */
