#ifndef DAQRAWTRIGGERRANGE_H
#define DAQRAWTRIGGERRANGE_H

#include <unistd.h>

#include "DAQDataFormats/RawDAQData.h"
#include "DAQDataFormats/TriggerDefines.h"

#define COPY_COPIES_INTERNAL

namespace daqdataformats {

  enum TriggerRangeMASKS{
    MASK_TRIGGER_LENGTH    = 0xFFFFFFFF,

    SHIFT_TRIGGER_LENGTH     = 0,
  };

  //! Class to hold the data from the global trigger range in the correct format
  /*!
    The RawTriggerRange contains the width of the trigger gate in
    500 nsec units.

    The class supports general initialization with an internal or external
    data buffer as well as copyin style initialization for use with
    STL containers.
  */
  class RawTriggerRange: public RawDAQData {
    
  public:
      
    RawTriggerRange();		                              //!< Constructor
    RawTriggerRange(void* input_buffer, uint32_t size=0); //!< Constructor with initialization (non-deep copy)
    RawTriggerRange(const RawTriggerRange& copy_in);          //! Copyin Constructor
    ~RawTriggerRange();                                       //!< Destructor
    

    bool copy(void* input_buffer, uint32_t size=0);       //!< Copy method from an external data block (non-deep)
    bool copy(const RawTriggerRange& copy_in);	          //!< Copy method from RawTriggerHeader format (deep copy)

    uint32_t getTriggerLength() const;        //! Get method for the trigger length

    bool setTriggerLength(uint32_t length);   //! Set method for the trigger length

    bool reset();                                         //!< Reset the class data
    bool clear();                                         //!< Clear the buffer
    bool resetBufferSource();	                          //!< Reset the class to use its internal buffer
    bool setBufferSource(void* input_buffer);             //!< Set the buffer to use for storage (can be external)

    bool isInternalBuffer() const;			  //!< Inquiry method for internal buffering type
            
    //! General write method that writes the raw Trigger Time formated data block to the target memory
    void* writeData(void* buffer);

    //! General read method that read the raw Trigger Time formated data block from the target memory
    void* readData(void* buffer);
    
    //! General sizeof method which returns the size of the Trigger Time formated data block in words (not bytes)
    uint32_t sizeofdata() const;

    void print() const;

    //! Assignment Operator
    bool operator=(const RawTriggerRange& copy_in);
    
  private:

    void init(void* input_buffer=0, uint32_t size=0); //!< Class initialization.  This method will zero associated buffers.

    //! The trigger time is a fixed size
    const static uint32_t TRIGRANGE_WORDS = 1;

    uint32_t _max_buffer_size;

    //! Raw block structure containing the packed information
    uint32_t  _trigRangeblock[TRIGRANGE_WORDS];

  };
  
  
} // END NAMESPCE daqdataformats
#endif /* DAQRAWTRIGGERRANGE_H */
