#ifndef DAQRAWTRIGGERTIME_H
#define DAQRAWTRIGGERTIME_H

#include <unistd.h>

#include "DAQDataFormats/RawDAQData.h"
#include "DAQDataFormats/TriggerDefines.h"

#define COPY_COPIES_INTERNAL	/* Define the behavior of copy constructors */

namespace daqdataformats {
  
  enum TriggerTimeMASKS{
    MASK_TIMESTAMP_LOW     = 0xFFFFFFFF,
    MASK_TIMESTAMP_HIGH    = 0x00FFFFFF,
    MASK_TIMESTAMP_SEC     = 0xFFFFFFFF,
    MASK_TIMESTAMP_USEC    = 0xFFFFFFFF,
    MASK_CLOCKSOURCE       = 0x000000FF,

    SHIFT_TIMESTAMP_LOW    = 0,
    SHIFT_TIMESTAMP_HIGH   = 0,
    SHIFT_TIMESTAMP_SEC    = 0,
    SHIFT_TIMESTAMP_USEC   = 0,
    SHIFT_CLOCKSOURCE      = 0,

    WORD_TIMESTAMP_LOW    = 0,
    WORD_TIMESTAMP_HIGH   = 1,
    WORD_TIMESTAMP_SEC    = 2,
    WORD_TIMESTAMP_USEC   = 3,
    WORD_CLOCKSOURCE      = 4,
  };

  //! Class to hold the data corresponding to the generated time of the trigger
  /*!
    The RawTriggerTime contains the information pertaining to the time at 
    which the trigger was generated, and what device generated the timestamp.

    The class supports general initialization with an internal or external
    data buffer as well as copyin style initialization for use with
    STL containers.
  */
  class RawTriggerTime: public RawDAQData {
    
  public:
      
    RawTriggerTime();		                            //!< Constructor
    RawTriggerTime(void* input_buffer, uint32_t size=0);    //!< Constructor with out data copy
    RawTriggerTime(const RawTriggerTime& copy_in);	    //!< Copyin Constructor (deep copy)
    ~RawTriggerTime();                                      //!< Destructor
    
    bool copy(void* input_buffer, uint32_t size=0); //!< Copy method from general buffer (non-deep copy)
    bool copy(const RawTriggerTime& copy_in);     //!< Copy method from RawTriggerTime format

    uint32_t getTriggerTimeStampLow() const;  //!< Get method for the low bits (0-31) of the trig timestamp
    uint32_t getTriggerTimeStampHigh() const; //!< Get method for the high bits (32+) of the trig timestamp
    uint64_t getTriggerTimeStamp() const;     //!< Get method for the full 64bits of the time stamp
    suseconds_t getTriggerTimeStampTvUsec() const;  //!< Get method for the low bits (0-31) of the trig timestamp timeval, corresponding to timeval.tv_usec
    time_t getTriggerTimeStampTvSec() const; //!< Get method for the high bits (32+) of the trig timestamp timeval, corresponding to timeval.tv_sec
    timeval getTriggerTimeStampTv() const;     //!< Get method for the full 64bits of the trig timestamp timeval struct
    uint32_t getClockSource() const;          //!< Get method for the time stamp source

    bool setTriggerTimeStampLow(uint32_t timestamp_low);      //!< Set method for the low bits (0-31) of the trig timestamp
    bool setTriggerTimeStampHigh(uint32_t timestamp_hi);      //!< Set method for the high bits (32+) of the trig timestamp
    bool setTriggerTimeStamp(uint64_t timestamp);       //!< Set method for 64bit time stamp
    bool setTriggerTimeStampTvUsec(suseconds_t timestamp_tv_usec);      //!< Set method for the low bits (0-31) of the trig timestamp timeval, corresponding to timeval.tv_usec
    bool setTriggerTimeStampTvSec(time_t timestamp_tv_sec);      //!< Set method for the high bits (32+) of the trig timestamp timeval, corresponding to timeval.tv_sec
    bool setTriggerTimeStampTv(timeval timestamp);       //!< Set method for 64bit trig timestamp timeval
    bool setClockSource(trigClockSource source);	          //!< Set method for trigger time clock source

    bool reset();                             //!< Reset the class data
    bool clear();                             //!< Clear the buffer
    bool resetBufferSource();                 //!< Reset the class to use its internal buffer
    bool setBufferSource(void* input_buffer); //!< Set the buffer to use for storage (can be external)

    bool isInternalBuffer() const;			  //!< Inquiry method for internal buffering type
            
    //! General write method that writes the raw Trigger Time formated data block to the target memory
    void* writeData(void* buffer);

    //! General read method that read the raw Trigger Time formated data block from the target memory
    void* readData(void* buffer);
    
    //! General sizeof method which returns the size of the Trigger Time formated data block in words (not bytes)
    uint32_t sizeofdata() const;

    //! Assignment Operator
    bool operator=(const RawTriggerTime& copy_in);


    void print() const;
    
  private:

    void init(void* input_buffer=0, uint32_t size=0); //!< Class initialization.  This method will zero associated buffers.

    //! The trigger time is a fixed size
    const static uint32_t TIMEBLOCK_WORDS = 5;

    //! Maximum size of the current buffer (used for external buffers)
    uint32_t _max_buffer_size;

    //! Raw block structure containing the packed information
    uint32_t  _trigTimeblock[TIMEBLOCK_WORDS];

  };
  
  
} // END NAMESPCE daqdataformats
#endif /* DAQRAWTRIGGERTIME_H */
