#include "DAQDataFormats/RawDAQData.h"

#include "DAQDataFormats/version.h"

#include <stdio.h>
 
using namespace daqdataformats;

//------------------------------------------------------------------------------
// Constructor
RawDAQData::RawDAQData()
    : major_version(PACKAGE_MAJOR_VERSION)
    , minor_version(PACKAGE_MINOR_VERSION)
    , _Buffer(NULL)  //Always nullify _Buffer in the constructor
    , _CRC32()
{
};

//------------------------------------------------------------------------------
// Addition to the default destructor
RawDAQData::~RawDAQData(){
        //free(_Buffer);		// Free the memory allocated for the data block
};


//------------------------------------------------------------------------------
// General I/O Methods
// These methods rely on the _Buffer pointer to the general data
// buffer being set by the class, and there being an implimentation
// of the sizeofdata() method.
// 
// These methods can be reimplimented by individual classes if their
// I/O is more complicated


// Method to read the data in from a raw buffer
void* RawDAQData::readData(void* buffer){
  uint32_t* read_buffer;
  uint32_t* data_buffer;

  data_buffer = (uint32_t*)_Buffer;

  if(buffer == 0){return 0;}

  read_buffer = (uint32_t*)buffer;

  for(uint32_t i = 0; i < this->sizeofdata(); ++i){
    data_buffer[i] = *read_buffer;
    read_buffer++;
  } // endfor

  return read_buffer;
}

// Method to write the data out to a file descriptor
readDataStatus RawDAQData::readData(int fd){
  if(fd>=0){
      ssize_t  ss;
      ss = read(fd,_Buffer,(size_t)this->sizeofdata()*sizeof(uint32_t));
      if (ss == -1 || ss == 0) return rds_error;
      if ((uint32_t)ss == this->sizeofdata()*sizeof(uint32_t)) return rds_complete;
      else                                           return rds_incomplete;
  }else{
    return rds_error;
  }
}


// Method to write the data out to a raw buffer
void* RawDAQData::writeData(void* buffer){
  uint32_t* write_buffer;
  uint32_t* source_buffer;

  source_buffer = (uint32_t*)_Buffer;

  if(buffer == 0){return 0;}

  write_buffer = (uint32_t*)buffer;

  for(uint32_t i = 0; i < this->sizeofdata(); ++i){
    *write_buffer = source_buffer[i];
    write_buffer++;
  } // endfor

  return write_buffer;
}

// Method to write the data out to a file descriptor
ssize_t RawDAQData::writeData(int fd){
  if(fd>0){
    return write(fd,_Buffer,(size_t)this->sizeofdata()*sizeof(uint32_t));
  }else{
    return -1;
  }
}


//------------------------------------------------------------------------------
void RawDAQData::printBuffer(){
  uint32_t* data_buffer;
  char* data_buffer2;
  data_buffer  = (uint32_t*)_Buffer;
  data_buffer2 = (char*)_Buffer;

  uint32_t cols = 4;

  for(uint32_t i = 0; i < this->sizeofdata(); ++i){

    // At the start of the line write the address offset
    if(i/cols){
      printf("%08xh |\t",(uint32_t)((data_buffer-(uint32_t*)_Buffer)*sizeof(uint32_t)));
    }

    // write the data word in hex
    printf("0x%08x\t",*data_buffer);
    ++data_buffer;

    // At the end of the line write the line in ASCII with suppression of
    // non printing characters
    if(!(i+1)/cols){
      printf("\t | ");
      for(uint32_t j=0; j < (sizeof(uint32_t)*cols); ++j){
	if( (*data_buffer2 < 32) || (*data_buffer2 > 126)){
	    printf(".");
	}else{
	  printf("%c",*data_buffer2);
	}
	  ++data_buffer2;
      }	// endfor j
      printf("\n");
    } // endif i
  } // endfor i
}

//------------------------------------------------------------------------------
void RawDAQData::printWord (uint32_t iword, bool is_bynary_printing) const {
   uint32_t* buffer_data = (uint32_t*)_Buffer;
   uint32_t value = SWAB32(buffer_data[iword]);

   //Check if to print in a bynary mode
    if(is_bynary_printing){
        const int SHIFT = 8 * sizeof( unsigned ) - 1;
        const unsigned MASK = 1 << SHIFT;

        for ( int i = 1; i <= SHIFT + 1; i++ ){
                std::cout << (  value & MASK ? '1' : '0' );
                value <<= 1;

                if ( i % 8 == 0 ) std::cout << ' ';
                }
        std::cout<<std::endl;
    } else {
        //Else print in hex mode
        printf("0x%08x",value);
   }// end of if(is_bynary_printing)

}//end of void printWord

//------------------------------------------------------------------------------
// General printing of a DAQDataFormat member. Should be overloaded.
void RawDAQData::print() const{

    std::cout<<"----------General printing of a DAQDataFormat member-----------size="<<this->sizeofdata()<<std::endl;
    std::cout<<"\nRAW:";
    for(uint32_t i = 0; i < this->sizeofdata(); ++i){
            printf(" ");
            this->printWord(i,false);
    }

    std::cout<<"\nRAW:";
    for(uint32_t i = 0; i < this->sizeofdata(); ++i){
            if(i > 0)std::cout<<"    ";
            this->printWord(i,true);
            std::cout<<"\n";
    }
}
//------------------------------------------------------------------------------
// Calculate Check Sum (CRC) for _Buffer.
// skip_last_words = number of words from the end to skip.
// Usually use just 0 or 1. 0 - to calculate CRC of the whole _Buffer.
// 0 to caclulate the CRC for the whole _Buffer
// 1 is usually used to append CRC to the _Buffer block
uint32_t RawDAQData::calculateCheckSum(int skip_last_words){

    _CRC32.reset();
    _CRC32.process_bytes((char*)_Buffer, (this->sizeofdata() - skip_last_words) * 4);
    //std::cout<<"\nCRC1 = "<<_CRC32.checksum();
    return _CRC32.checksum();

/*
    _CRC32.reset();
    printf("\n\n\n");
    char* jj = (char*)_Buffer;
    //uint32_t* jj = (uint32_t*)_Buffer;
    for(int i = 0; i < (this->sizeofdata() - skip_last_words) * 4; ++i){
    //for(int i = 0; i < this->sizeofdata() - skip_last_words; ++i){
        _CRC32.process_bytes(jj, 1);
        printf("%u, crc %u|",*jj,_CRC32.checksum());
        jj++;
        }
    std::cout<<" CRC2 = "<<_CRC32.checksum();
 return _CRC32.checksum();
*/
}

//------------------------------------------------------------------------------
// Add the CRC into the end of fBuffer. CRC of all but the last words of the _Buffer
void RawDAQData::addCheckSum(){
 uint32_t* buffer_data = (uint32_t*)_Buffer;
 uint32_t crc = this->calculateCheckSum(1);

 buffer_data[this->sizeofdata() -1] = SWAB32(crc);
}


//------------------------------------------------------------------------------
// Get the pointer to the internal _Buffer. Use this cautiously.
void* RawDAQData::getBuffer() const {
    return _Buffer;
}



