#include "DAQDataFormats/RawEvent.h"
#include "DAQDataFormats/DataFormatException.h"
#include <boost/lexical_cast.hpp>
#include <cstdio>

#ifndef RAWBUFF32
// Define a macros to make accessing the raw buffer easier
#define RAWBUFF32 ((uint32_t*)this->_Buffer)
#define RAWBUFF16 ((uint16_t*)this->_Buffer)
#endif


using namespace daqdataformats;

//------------------------------------------------------------------------------
// Default constructor
RawEvent::RawEvent(uint32_t size){

    /// Check the size of the buffer
    this->checkSize(size);

    /// Set the maximum size
    _max_buffer_size = size;

    _InternalBuffer.clear();
    _InternalBuffer.resize(_max_buffer_size, 0);

    /// Set pointers to point to the internal buffer
    this->resetBufferSource();

    /// Initialize the event (header)
    this->init();
}

//------------------------------------------------------------------------------
RawEvent::RawEvent(const RawEvent& copy_in){
    this->copy(copy_in);
}

//------------------------------------------------------------------------------
RawEvent::RawEvent(void* rawevent_buffer, uint32_t size){

    /// Check the size of the buffer
    this->checkSize(size);

    /// Set the maximum size
    _max_buffer_size = size;

    // Initialize the buffer with a size and set them to zero
    _InternalBuffer.clear();
    _InternalBuffer.resize(_max_buffer_size, 0);


    /// Set pointers to point to the internal buffer
    this->resetBufferSource();

    /// Initialize the event (header)
    this->init();

    // Now, setup the pointer to the external buffer
    this->setBufferSource(rawevent_buffer);
}

//------------------------------------------------------------------------------
// Destructor
RawEvent::~RawEvent() { }


//------------------------------------------------------------------------------
// Check size
bool RawEvent::checkSize(uint32_t size) const {

    if (size < RAWEVENTHEADER_SIZE + RAWEVENTTAIL_SIZE + 1) {
            std::string msg("An external input buffer was given to the ");
            msg.append("RawEvent constructor that was too small (size=");
            msg.append(boost::lexical_cast<std::string>(size) + ", minimum size=");
            msg.append(boost::lexical_cast<std::string>(RAWEVENTHEADER_SIZE + RAWEVENTTAIL_SIZE + 1) + ").");
            GENERATE_DATAFORMAT_EXCEPTION(msg);
        }

    return true;
}

//------------------------------------------------------------------------------
// Initialization of Data members
void RawEvent::init(){
    
    _Header.init();
    // Add the _Trigger size to the RawEvent Header
    _Header.setEventSize(_Header.sizeofdata() + _Trigger.sizeofdata());

    _Trigger.getTriggerHeader()->setDelimiter(TriggerHeaderDelimiter);
    

    // Initialize the _FloatingDataBlockNumber
    _FloatingDataBlockNumber = -1;
}

//------------------------------------------------------------------------------
void RawEvent::copy(const RawEvent& copy_in){


    _max_buffer_size = copy_in._max_buffer_size;
    _InternalBuffer.reserve((long unsigned int)_max_buffer_size);

    /// Copy the data from one buffer to the other
    for(uint32_t i = 0; i < _max_buffer_size; ++i)
        _InternalBuffer[i] = copy_in._InternalBuffer[i];

    // Setup the main buffer pointer
    // If the copy_in's buffer is external, point to that external memory
    if(copy_in.isInternalBuffer()) this->resetBufferSource();
    else                           this->setBufferSource(copy_in._Buffer);

    // Setup the FloatingDataBlock
    _FloatingDataBlockNumber = copy_in._FloatingDataBlockNumber;
    if(_FloatingDataBlockNumber > -1)this->setFloatingDataBlock(copy_in._FloatingDataBlockNumber);
}


//------------------------------------------------------------------------------
// Add a MicroSlice into the RawEvent changing the Header appropriately
// The method should be identical to RawMicroSlice::readData() (except for not reading the MicroSlice Timing marker)
// If check_mc_flag = true, the method checks whether the DataBlock is MC generated
// and adds an appropriate flag to the Event Header.
void RawEvent::addDataBlock(void* datablock_buffer, bool check_mc_flag){

    // Set the floating datablock here
    _FloatingDataBlock.setBufferSource(&RAWBUFF32[this->sizeofdata()]);

    _FloatingDataBlock.readData(datablock_buffer);

    // Setup the RawEventHeader Header
    // Advance the RawEvent DataBlock number and the size
    _Header.addEventSize(_FloatingDataBlock.sizeofdata());
    _Header.advanceDataBlockNumber();

    // Check if the DataBlock is the MC generated
    // If it is, set the appropriate MicroSlice MC flag
    if(!check_mc_flag)return;
    this->setMonteCarloFlag(_FloatingDataBlock.getMonteCarloFlag());
}

//------------------------------------------------------------------------------
// Add a MicroSlice of a known size into the RawEvent changing the RawEvent Header appropriately
// The method should be identical to RawMicroSlice::readData() (except for not reading the MicroSlice Timing marker)
// If check_mc_flag = true, the method checks whether the DataBlock is MC generated
// and adds an appropriate flag to the Event Header.
void RawEvent::addDataBlock(void* datablock_buffer, uint32_t datablock_size, bool check_mc_flag){

    uint32_t Buffer_current_size = this->sizeofdata(); //remember current buffer position

    // Setup the RawEventHeader Header
    // Advance the RawEvent DataBlock number and the size
    _Header.addEventSize(datablock_size);
    _Header.advanceDataBlockNumber();

    //Add the MicroSlice internal buffer to the RawEvent _Buffer starting from the last _Buffer size position
    this->readData(datablock_buffer, datablock_size, Buffer_current_size);

    // Check if the DataBlock is the MC generated
    // If it is, set the appropriate MicroSlice MC flag
    if(!check_mc_flag)return;
    RawDataBlock datablock(datablock_buffer);
    this->setMonteCarloFlag(datablock.getMonteCarloFlag());
}

//------------------------------------------------------------------------------
// Usually just packs the Header leaving the rest of the _Buffer unchanged.
// Also, puts the Header in the tail as well. Puts CRC in the very end.
void RawEvent::close(bool produce_crc){

    // If CRC is calculated, set the isCRCCalculationUsed flag
    if(produce_crc)_Header.setCRCCalculationUsed(true);

    // Add the size of the tail + CRC
    _Header.addEventSize(RAWEVENTTAIL_SIZE + 1);

    //Now Setup the tail
    _Tail.setBufferSource(&RAWBUFF32[this->sizeofdata() - RAWEVENTTAIL_SIZE - 1]);
    _Tail.init();
    _Tail.setEventNumberLo(_Header.getEventNumberLo());
    _Tail.setEventNumberHi(_Header.getEventNumberHi());
    _Tail.setEventSize    (_Header.getEventSize());

    // Finally, add CRC in the very end of _Buffer
    if(produce_crc)this->addCheckSum();
}

//------------------------------------------------------------------------------
// This is a general method to read the datablock from RawEvent_buffer to _Buffer
void* RawEvent::readData(void* RawEvent_buffer){
    //Read the header and unpack it to find out the size of the RawEvent
    RawEvent_buffer = _Header.readData(RawEvent_buffer);

    //Continue to read
    return this->readData(RawEvent_buffer, this->sizeofdata() - _Header.sizeofdata(), _Header.sizeofdata());
}

//------------------------------------------------------------------------------
void* RawEvent::readData(void* void_buffer, uint32_t size){

  uint32_t* buffer = (uint32_t*)void_buffer;

  if(buffer == 0){return 0;}

  for(uint32_t i = 0; i < size; ++i )
    RAWBUFF32[i] = buffer[i];

  // Return the address of the buffer advanced by 1 word;
  return (void*)(buffer + size);
}

//------------------------------------------------------------------------------
void* RawEvent::readData(void* buffer_void, uint32_t size, uint32_t _Buffer_start_position){

  uint32_t* buffer = (uint32_t*)buffer_void;

  if(buffer == 0){return 0;}

  for(uint32_t i = 0; i < size; ++i )
    RAWBUFF32[i+_Buffer_start_position] = buffer[i];

  // Return the address of the buffer advanced by 1 word;
  return (void*)(buffer+size);
}

//------------------------------------------------------------------------------
//Getters
    RawEventHeader* RawEvent::getHeader()                  { return &_Header; }
    RawEventTail*   RawEvent::getTail() {
        

        uint32_t* pointer = ((uint32_t*)(this->getBuffer())) + this->sizeofdata() - 1 - RAWEVENTTAIL_SIZE;
        _Tail.setBufferSource(pointer);

        return &_Tail;
    }
    RawTrigger*     RawEvent::getTrigger()                 { return &_Trigger; }
    RawDataBlock*   RawEvent::getFloatingDataBlock()       { return &_FloatingDataBlock; }
    uint32_t        RawEvent::getHeaderMarkerLeft()  const { return _Header.getMarkerLeft();}
    uint32_t        RawEvent::getHeaderMarkerRight() const { return _Header.getMarkerRight();}
    uint32_t        RawEvent::getVersion()           const { return _Header.getVersion();}
    uint32_t        RawEvent::getDataLoggerID()      const { return _Header.getDataLoggerID();}
    uint32_t        RawEvent::getRunNumber()         const { return _Header.getRunNumber();}
    uint32_t        RawEvent::getRunType()           const { return _Header.getRunType();}
    uint32_t        RawEvent::getConfigIDX()         const { return _Header.getConfigIDX();}
    uint32_t        RawEvent::getPartition()         const { return _Header.getPartition();}
    uint32_t        RawEvent::getSubrun()            const { return _Header.getSubrun();}
    uint32_t        RawEvent::getEventNumberLo()     const { return _Header.getEventNumberLo();}
    uint32_t        RawEvent::getEventNumberHi()     const { return _Header.getEventNumberHi();}
    uint32_t        RawEvent::getEventType()         const { return _Header.getEventType();}
    uint32_t        RawEvent::getEventSize()         const { return _Header.getEventSize();}
    uint32_t        RawEvent::getDataBlockNumber()   const { return _Header.getDataBlockNumber();}
    bool            RawEvent::isEventIncomplete()    const { return _Header.isEventIncomplete();}
    bool            RawEvent::getMonteCarloFlag()    const { return _Header.getMonteCarloFlag();}
    bool            RawEvent::isCRCCalculationUsed() const { return _Header.isCRCCalculationUsed();}
    uint64_t  RawEvent::getEventNumber() const {
      uint64_t evnt;
      uint32_t low, high;

      low  = this->getEventNumberLo();
      high = this->getEventNumberHi();

      evnt=(uint64_t)high<<32 | low;
      return evnt;
    }

uint32_t RawEvent::getFreeMemory() const {
            if(this->sizeofdata() > _max_buffer_size)return 0;
            return _max_buffer_size - this->sizeofdata();
}

//------------------------------------------------------------------------------
//Setters

    bool RawEvent::setTrigger       (RawTrigger&  trigger){
        return this->setTrigger(trigger.getBuffer());
    }
    bool RawEvent::setTrigger       (void*  pointer){
        _Trigger.readData(pointer);
        return true;
    }
    bool RawEvent::setVersion       (uint32_t inp) {return _Header.setVersion       (inp);}
    bool RawEvent::setDataLoggerID  (uint32_t inp) {return _Header.setDataLoggerID  (inp);}
    bool RawEvent::setRunNumber     (uint32_t inp) {return _Header.setRunNumber     (inp);}
    bool RawEvent::setRunType       (uint32_t inp) {return _Header.setRunType       (inp);}
    bool RawEvent::setConfigIDX     (uint32_t inp) {return _Header.setConfigIDX     (inp);}
    bool RawEvent::setPartition     (uint32_t inp) {return _Header.setPartition     (inp);}
    bool RawEvent::setSubrun        (uint32_t inp) {return _Header.setSubrun        (inp);}
    bool RawEvent::setEventNumberLo (uint32_t inp) {return _Header.setEventNumberLo (inp);}
    bool RawEvent::setEventNumberHi (uint32_t inp) {return _Header.setEventNumberHi (inp);}
    bool RawEvent::setEventType     (uint32_t inp) {return _Header.setEventType     (inp);}
    bool RawEvent::setEventIncomplete(bool tf)     {return _Header.setEventIncomplete(tf);}
    bool RawEvent::setMonteCarloFlag (bool tf)     {return _Header.setMonteCarloFlag(tf);}

    bool RawEvent::setEventNumber(uint64_t evnt) {
      uint32_t low, high;

     low  = evnt & 0xffffffff;
     high = evnt >> 32;

     this->setEventNumberHi(high);
     this->setEventNumberLo(low);

     return true;
    }

    bool RawEvent::setBufferSource(void* pointer){
        _Buffer = pointer;
        _Header.setBufferSource(pointer);

        uint32_t* pp = (uint32_t*)pointer + _Header.sizeofdata();
        _Trigger.setBufferSource(pp);


        // reset the Floating MicroSlice number
        _FloatingDataBlockNumber = -1;
        return true;
    }
    bool RawEvent::resetBufferSource(){
        this->setBufferSource(&_InternalBuffer[0]);
        return true;
    }

// clears all data in the buffer
    bool RawEvent::clear(){
        int max_data = this->sizeofdata();

        // Loop over the buffer and zero it
        for(int i=0; i< max_data; ++i)
            RAWBUFF32[i]=0;

        return true;
    }
    bool RawEvent::reset(){
        this->resetBufferSource();
        this->clear();
        return true;
    }
    bool RawEvent::isInternalBuffer() const {

        if((this->_Buffer) == (void*)(&_InternalBuffer[0]))
            return true;
        return false;
    }

    //------------------------------------------------------------------------------
    bool RawEvent::setFloatingDataBlock(uint32_t idatablock){
        if(idatablock == 0){
            _FloatingDataBlockNumber = idatablock;
            return _FloatingDataBlock.setBufferSource(&RAWBUFF32[_Header.sizeofdata() + _Trigger.sizeofdata()]);
            }

        if((int)idatablock == _FloatingDataBlockNumber)return true;

        if((int)idatablock > _FloatingDataBlockNumber && _FloatingDataBlockNumber > -1){
            uint32_t* pointer = (uint32_t*)_FloatingDataBlock.getBuffer();
            for(; _FloatingDataBlockNumber < (int)idatablock; _FloatingDataBlockNumber++){
                 pointer += _FloatingDataBlock.sizeofdata();
                _FloatingDataBlock.setBufferSource(pointer);
            }
            return true;
        }

        // Firrst point to the first DataBlock
        uint32_t* pointer = (uint32_t*)&RAWBUFF32[_Header.sizeofdata() + _Trigger.sizeofdata()];
        _FloatingDataBlock.setBufferSource(pointer);

        for(_FloatingDataBlockNumber = 0; _FloatingDataBlockNumber < (int)idatablock; _FloatingDataBlockNumber++){
            pointer += _FloatingDataBlock.sizeofdata();
            _FloatingDataBlock.setBufferSource(pointer);
            }
        return true;
    }

void RawEvent::nextEvent(){
    
    // Advance event number
    this->getHeader()->advanceEventNumber();
    //reset DataBlock number
    this->getHeader()->setDataBlockNumber(0);
    //reset the size
    this->getHeader()->setEventSize(_Trigger.sizeofdata() + _Header.sizeofdata());

    // reset the _FloatingDataBlockNumber
    _FloatingDataBlockNumber = -1;

}




//------------------------------------------------------------------------------
// Printing the RawEvent structure
void RawEvent::print() {
    printf("---RawEvent---------------------------------------------size = %u\n",this->sizeofdata());
    _Header.print();


    /// Print the Trigger Block

    _Trigger.print();

    /// Print all DataBlocks
    for(uint32_t idatablock = 0; idatablock < this->getDataBlockNumber(); ++idatablock){
        this->setFloatingDataBlock(idatablock);
        this->getFloatingDataBlock()->print();
    }





    std::cout<<"\nRAW:";
    for(uint32_t i = 0; i < this->sizeofdata(); ++i){
        this->printWord(i,false);
        printf(" ");
    }

    std::cout<<"\nRAW:";
    for(uint32_t i = 0; i < this->sizeofdata(); ++i){
            if(i > 0)std::cout<<"    ";
            this->printWord(i,true);
            std::cout<<"\n";
            }
 
    printf("---end of RawEvent---------------------------------------size = %u\n",this->sizeofdata());

}

//------------------------------------------------------------------------------
// Size of the RawEvent _Buffer in words
uint32_t RawEvent::sizeofdata()const  { return _Header.getEventSize(); }
